/********************************** <lalVerbatim file="PowerApplySearchCV">
Author: Brady, P
$Id: ApplySearch.c,v 1.15 2001/08/30 16:55:11 patrick Exp $
**************************************************** </lalVerbatim> */

#include <lal/LALStdlib.h>
#include <lal/LALDatatypes.h>
#include <lal/ExcessPower.h>
#include <lal/TFTransform.h>
#include <lal/Comm.h>
#include <lal/AVFactories.h>
#include "ApplySearch.h"
#include "EPExch.h"

NRCSID( APPLYSEARCHC, "power $Id: ApplySearch.c,v 1.15 2001/08/30 16:55:11 patrick Exp $");

void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *applyParams
    )
{
  COMPLEX8FrequencySeries *fseries;
  EPSearchParams          *params = NULL;
  INT4                     myRank;
  RealDFTParams           *dftparams=NULL;
  LALWindowParams          winParams;
  InitExchParams           initExchParams;
  TFTile                  *tfTile=NULL;

  INITSTATUS (status, "LALApplySearch", APPLYSEARCHC);
  ATTATCHSTATUSPTR (status);

  /* check input, output and parameters */
  ASSERT (output, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL);
  ASSERT ( !(output->result), status, APPLYSEARCHH_ENNUL, APPLYSEARCHH_MSGENNUL);
  
  ASSERT (applyParams, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL);
  ASSERT (applyParams->mpiParams, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL);
  ASSERT (applyParams->searchParams, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL);
  
  /* cast the parameter structure, assign MPI information */
  params = (EPSearchParams *) applyParams->searchParams;
  winParams.type=Rectangular;
  winParams.length=params->ntotT;

  
  /* 
   *
   * if search master, get the initial number of nodes in the search comm 
   *
   */

  
  if ( params->searchMaster )
  {
    /* only do this the first time the master is called */
    if ( !params->numSlaves )
    {
      /* allocate memory and check for success */
      params->numSlaves = (UINT4 *) LALMalloc( sizeof(UINT4) );
      memset( params->numSlaves, 0, sizeof(UINT4) );
      if( !params->numSlaves )
      {
        ABORT( status, APPLYSEARCHH_EALOC, APPLYSEARCHH_MSGEALOC );
      }

      /* get the number of search slaves */
      MPI_Comm_size( *(applyParams->mpiParams->comm), params->numSlaves );
      *(params->numSlaves) -= 1;
    }
    /* set the number of events to zero */
    params->numEvents = 0;
  }

  
  /*
   *
   * determine rank and check that we're running on a physical slave
   *
   */

  
  MPI_Comm_rank( *(applyParams->mpiParams->comm), &myRank );

  /* check that the master is running on node rank zero */
  if ( params->searchMaster && myRank != 0 )
  {
    ABORT( status, APPLYSEARCHH_ERANK, APPLYSEARCHH_MSGERANK );
  }

  /* and that the slave is not running on node rank zero */
  if ( !params->searchMaster && myRank == 0 )
  {
    ABORT( status, APPLYSEARCHH_ERANK, APPLYSEARCHH_MSGERANK );
  }
 
  initExchParams.mpiComm   = *(applyParams->mpiParams->comm);
  initExchParams.myProcNum = myRank;

  
  /* 
   *
   * the main engine starts here.  On the master 
   *
   */

  
  if ( params->searchMaster )
  {
  
    /* 
     *
     * if there are slaves
     *
     */

    if ( *(params->numSlaves) )
    {
      ExchParams *thisExch = NULL;

      /* wait for a message */
      LALInitializeExchange (status->statusPtr, &thisExch, NULL, &initExchParams);
      CHECKSTATUSPTR (status);
      
      /* case the exchange type */
      switch (thisExch->exchObjectType)
      {
        
        /*
         *
         * how many segments slave should analyze
         *
         */
        
        case ExchDataList:

          /*
           *
           * if there is more data to analyze
           *
           */
          
          if ( params->currentSegment < params->initParams->numSegments )
          {
            UINT4       tmpDutyCycle;

            /* make sure there is data to analyze */
            for ( tmpDutyCycle=0 ; tmpDutyCycle < params->initParams->segDutyCycle && 
                params->currentSegment < params->initParams->numSegments ; tmpDutyCycle++ )
            {
            }
            
            /* tell the slave how many segments */
            LALExchangeUINT4 (status->statusPtr, &tmpDutyCycle, thisExch);
            CHECKSTATUSPTR (status);
            
            /* tell the slave the first segment */
            LALExchangeUINT4 (status->statusPtr, &(params->currentSegment), thisExch);
            CHECKSTATUSPTR (status);

            /* increment to the next segment number to be analyzed */
            params->currentSegment += tmpDutyCycle;
          }

          /*
           *
           * otherwise tell the slave there is no more data to analyze 
           *
           */
          
          else
          {
            UINT4       tmpDutyCycle;

            
            tmpDutyCycle = 0;
            LALExchangeUINT4 (status->statusPtr, &tmpDutyCycle, thisExch);
            CHECKSTATUSPTR (status);
          }
          break;
  
        /*
         *
         * exchange event list
         *
         */

        case ExchEPEvent:
          /* get the number of events */
          LALExchangeUINT4 (status->statusPtr, &(params->numEvents), thisExch);
          CHECKSTATUSPTR (status);

          /* if there are events */
          if ( params->numEvents >0 )
          {
            PowerOutPutParams  outPutParams;

            /* set outPutParams */
            outPutParams.numEvents=params->numEvents;
            outPutParams.flow=params->tfTilingInput->flow;

            /* get the start time of the data segment in GPS nanoseconds */
            LALExchangeINT8 (status->statusPtr, &(outPutParams.tstart), thisExch);
            CHECKSTATUSPTR (status);

            /* get the event list and stick them into temporary memory */
            LALExchangeEPEventList(status->statusPtr, &(tfTile), 
                params->numEvents, thisExch);
            CHECKSTATUSPTR (status);
            /* build the wrapper's output structure */
            BuildPowerOutPut (status->statusPtr, output, tfTile, &outPutParams);
            CHECKSTATUSPTR (status);
            /* free the temporary memory containin the events */
            while (tfTile)
            {
              TFTile *thisTile;
              thisTile = tfTile;
              tfTile = tfTile->nextTile; 
              LALFree( thisTile );
            }
          }
          /* if there are no events set output to zero */
          else
          {
            output->numOutput     = 0;
            output->result        = NULL;
          }
          break;
      

        /*
         *
         *  a slave is finished
         *
         */

        case ExchFinished:
          --*(params->numSlaves);
          break;


        /*
         *
         * unrecognized type
         *
         */

        default: 
          ABORT( status, APPLYSEARCHH_EUEXT, APPLYSEARCHH_MSGEUEXT );
      }
      /* end case exhange type */

      /* finish the exchange */
      LALFinalizeExchange (status->statusPtr, &thisExch);
      CHECKSTATUSPTR (status);

      /* 
       *
       * set fraction remaining
       *
       */

      output->fracRemaining = 1.0 - 
        (REAL4)(params->currentSegment - *(params->numSlaves)) / 
        (REAL4)(params->initParams->numSegments);
      /* BEGIN KLUDGE */
      if ( output->fracRemaining < 0.001 ) { output->fracRemaining = 0.01; } 
      /* END KLUDGE */
    }

    /*
     *
     * if there are no slaves left
     *
     */

    else
    {
      output->fracRemaining = 0;
      output->notFinished   = 0;
    }
  }

  /*
   *
   * if I'm a slave
   *
   */

  else
  {
    UINT4                   tmpDutyCycle;
    INT4                    i,j;
    ExchParams              exchEvents;
    ExchParams              exchDutyCycle;
    ExchParams              exchFinished;
    ExchParams             *thisExch = NULL;
    EPDataSegment          *dummySegment = NULL;
    

    /* exchange finished message */
    exchFinished.exchObjectType  = ExchFinished;
    exchFinished.send            = 0;  /* master sends */
    exchFinished.numObjects      = 0;
    exchFinished.partnerProcNum  = 0;  /* master */

    /* exchange duty cycle information */
    exchDutyCycle.exchObjectType = ExchDataList;
    exchDutyCycle.send           = 0; /* master sends */
    exchDutyCycle.numObjects     = 1; /* i send */
    exchDutyCycle.partnerProcNum = 0; /* master */

    /* exchange event information */
    exchEvents.exchObjectType = ExchEPEvent;
    exchEvents.send           = 1; /* i send */
    exchEvents.numObjects     = 1; /* i send */
    exchEvents.partnerProcNum = 0; /* master */


    /*
     *
     * find out if there is more data to analyze
     *
     */
    
    
    LALInitializeExchange (status->statusPtr, &thisExch, 
        &exchDutyCycle, &initExchParams);
    CHECKSTATUSPTR (status);

    LALExchangeUINT4 (status->statusPtr, &tmpDutyCycle, thisExch);
    CHECKSTATUSPTR (status);

    /*
     *
     * if there are no segments,  tell the master we're outta here!
     *
     */
    
    if ( !tmpDutyCycle )
    {

#ifdef POWER_SO_DEBUG
      fprintf( stderr, "search slave: shutting down\n" );
      fflush( stderr );
#endif

      LALFinalizeExchange (status->statusPtr, &thisExch);
      CHECKSTATUSPTR (status);

      /* this is the slave shutdown handshake */
      LALInitializeExchange (status->statusPtr, &thisExch, 
          &exchFinished, &initExchParams);
      CHECKSTATUSPTR (status);

      LALFinalizeExchange (status->statusPtr, &thisExch);
      CHECKSTATUSPTR (status);

      /* make sure to tell the wrapperAPI that we're done */
      output->notFinished   = 0;

      DETATCHSTATUSPTR (status);
      RETURN (status);

    }

    /*
     *
     * if there are segments,  find out the first one to analyze
     *
     */

    else
    {
      LALExchangeUINT4 (status->statusPtr, &(params->currentSegment), thisExch);
      CHECKSTATUSPTR (status);
    }

    LALFinalizeExchange (status->statusPtr, &thisExch);
    CHECKSTATUSPTR (status);

    /* assign temporary memory for the frequency data */
    fseries = (COMPLEX8FrequencySeries *) LALMalloc (sizeof(COMPLEX8FrequencySeries));
    strncpy( fseries->name, "anonymous", LALNameLength * sizeof(CHAR) );
    fseries->data = NULL;
    LALCCreateVector (status->statusPtr, &fseries->data, 
        params->initParams->numPoints/2 + 1);
    CHECKSTATUSPTR (status);


    /* create the dft params */
    LALCreateRealDFTParams(status->statusPtr , &dftparams, &winParams, 1);
    CHECKSTATUSPTR (status);

    /* point dummySegment to the first segment to analyze */
    dummySegment = params->epSegVec->data + params->currentSegment;

    /*
     *
     * loop over data applying excess power method
     *
     */
    
    for ( i=0 ; i<tmpDutyCycle ; i++)
    {

#ifdef POWER_SO_DEBUG
      fprintf( stderr, "search slave %d: filtering segment %d\n", 
          myRank, params->currentSegment);
      fflush( stderr );
#endif

      /* compute the DFT of input time series */
      LALComputeFrequencySeries (status->statusPtr, fseries, 
          dummySegment->data, dftparams);
      CHECKSTATUSPTR (status);

      /* check that deltaF agrees with that of response */
      if ( fabs( dummySegment->spec->deltaF - fseries->deltaF ) > 0.000001 )
      {
        ABORT (status, APPLYSEARCHH_EDELF, APPLYSEARCHH_MSGEDELF );
      }
#ifdef POWER_SO_DEBUG
      fprintf( stderr, "search slave: first element %d, %f \n", 
          dummySegment->number, (dummySegment->data->data->data)[100]);
      fflush( stderr );
#endif
      /* normalize the data stream so that rms of Re or Im is 1 */
      for (j=0 ; j<fseries->data->length ; j++)
      {
        REAL4 tmpVar = sqrt( 4 * dummySegment->data->deltaT / 
            dummySegment->spec->data->data[j] );
        fseries->data->data[j].re *= tmpVar;
        fseries->data->data[j].im *= tmpVar;
      }

      /* create time-frequency tiling of plane */
      params->tfTilingInput->deltaF=fseries->deltaF;
      LALCreateTFTiling (status->statusPtr, &(params->tfTiling), params->tfTilingInput);
      CHECKSTATUSPTR (status);

      /* compute the TFplanes for the data segment */
      LALComputeTFPlanes (status->statusPtr, params->tfTiling, fseries);
      CHECKSTATUSPTR (status);

      /* search these planes */
      LALComputeExcessPower (status->statusPtr, params->tfTiling, params->compEPInput);
      CHECKSTATUSPTR (status);

      /* compute the likelihood for slightly better detection method */
      LALComputeLikelihood  (status->statusPtr, &(params->lambda), params->tfTiling);
      CHECKSTATUSPTR (status);

      /* sort the results */
      LALSortTFTiling (status->statusPtr, params->tfTiling);
      CHECKSTATUSPTR (status);

      /* count the number of events  */
      /* change alphaThreshold to match with confidence */
      LALCountEPEvents(status->statusPtr, &(params->numEvents), 
          params->tfTiling, params->alphaThreshold);
      CHECKSTATUSPTR (status);

      /* Don't send more than 16 events to the Master */
      if ( params->numEvents > params->events2Master )
      {
        params->numEvents = params->events2Master;
      }

      /* Tell master how many events to expect */
      LALInitializeExchange (status->statusPtr, &thisExch, 
          &exchEvents, &initExchParams);
      CHECKSTATUSPTR (status);

      LALExchangeUINT4 (status->statusPtr, &(params->numEvents), thisExch);
      CHECKSTATUSPTR (status);

      if ( params->numEvents > 0 )
      {
        INT8 tstartNS=0;

        /* convert epoch to GPS nanoseconds */
        tstartNS  = 1000000000L * 
            (INT8) dummySegment->data->epoch.gpsSeconds;
        tstartNS += (INT8) dummySegment->data->epoch.gpsNanoSeconds;

        /* Send master start time of segment in GPS nanoseconds */
        LALExchangeINT8 (status->statusPtr, &(tstartNS), thisExch);
        CHECKSTATUSPTR (status);

        /* send the master the event list */
        LALExchangeEPEventList(status->statusPtr, &(params->tfTiling->firstTile), 
            params->numEvents, thisExch);
        CHECKSTATUSPTR (status);
      }

      LALFinalizeExchange (status->statusPtr, &thisExch);
      CHECKSTATUSPTR (status);

      /* destroy time-frequency tiling of planes */
      LALDestroyTFTiling (status->statusPtr, &(params->tfTiling));
      CHECKSTATUSPTR (status);

    }

    /* destroy the dftparams for computing frequency series */
    LALDestroyRealDFTParams (status->statusPtr, &dftparams);
    CHECKSTATUSPTR (status);

    /* destroy temporary memory for the frequency data */
    LALCDestroyVector (status->statusPtr, &fseries->data);
    CHECKSTATUSPTR (status);
    LALFree(fseries);

    /* Dont know if this is right */
    output->numOutput     = 0;
    output->result        = NULL;
    output->fracRemaining = 0;
    output->notFinished   = 1;

  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

