/***************************** <lalVerbatim file="StochasticApplySearchCV">
Author: Anderson, W.G., Romano, J.D., and Whelan, J.T.
$Id: StochasticApplySearch.c,v 1.11 2001/08/27 03:50:00 whelan Exp $
 **************************************************** </lalVerbatim> */

/******************************* <lalLaTeX file="StochasticApplySearchC">
  \subsection{Module \texttt{StochasticApplySearch.c}}
  \label{s:StochasticApplySearch.c}

  The LAL-wrapper interface function for ...

  \subsection*{Description}

  \subsection*{Algorithms}

  \subsection*{Uses}

  \subsection*{Notes}

  \vfill{\footnotesize\input{StochasticApplySearchCV}}

 *************************************************** </lalLaTeX> */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <lal/LALStdlib.h>
#include <lal/LALDatatypes.h>
#include <lal/Comm.h>
#include <lal/AVFactories.h>
#include <lal/TimeFreqFFT.h>
#include <lal/CoarseGrainFrequencySeries.h>
#include "StochasticApplySearch.h"

NRCSID (STOCHASTICAPPLYSEARCHC, "$Id: StochasticApplySearch.c,v 1.11 2001/08/27 03:50:00 whelan Exp $");

void
LALApplySearch(
      LALStatus             *status,
      LALSearchOutput       *output,
      LALSearchInput        *input,
      LALApplySearchParams  *applyParams
              )
{

   INT4                                myRank;

   COMPLEX8FrequencySeries             fseries1;
   COMPLEX8FrequencySeries             fseries2;
   COMPLEX8FrequencySeries             ccSpectrum;
   COMPLEX8FrequencySeries             ccSpectrumOut;
   StochasticCrossCorrelationInput     ccInput;
   FrequencySamplingParams             cgParams;
   gpsTimeInterval                     timeInterval;
   UINT8                               segLengthNanoSeconds;
   UINT8                               segStartNanoSeconds;
   UINT8                               segStopNanoSeconds;

   StochasticSearchParams*             params = NULL;
   StochasticDataSegment*              dummySegment = NULL;

   INITSTATUS (status, "LALApplySearch", STOCHASTICAPPLYSEARCHC);
   ATTATCHSTATUSPTR (status);


   /* 
    *
    * check output and parameters 
    * 
    */

   ASSERT (output, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);

   ASSERT (!(output->result), 
         status, 
         STOCHASTICAPPLYSEARCHH_ENNUL, 
         STOCHASTICAPPLYSEARCHH_MSGENNUL);

   ASSERT (applyParams, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);

   ASSERT (applyParams->mpiParams, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);

   ASSERT (applyParams->searchParams, 
         status, 
         STOCHASTICAPPLYSEARCHH_ENULL, 
         STOCHASTICAPPLYSEARCHH_MSGENULL);


   /* cast the parameter structure */ 

   params = (StochasticSearchParams *) applyParams->searchParams;


   /* 
    *
    * perform the search
    *
    */

   /* 
    * 
    * NOTE: WE DO THE ANALYSIS ON **ONLY** THE SEARCH MASTER SINCE
    * WE CONSIDER THE SIMPLE CASE OF A SEARCH FOR A STOCHASTIC
    * BACKGROUND SIGNAL WITH OMEGA_GW=CONST.  
    * SEARCHES FOR STOCHASTIC SIGNALS HAVING DIFFERENT SPECTRA CAN 
    * BE DONE ON MULTIPLE SEARCH SLAVES.
    * THIS IS WORK FOR THE FUTURE!!
    * 
    */   

   /* if not search master, abort */
   if (!(params->searchMaster))
   {
      ABORT (status, 
            STOCHASTICAPPLYSEARCHH_EMSTR, 
            STOCHASTICAPPLYSEARCHH_MSGEMSTR);
   }

   /* determine rank and check that search master is running on node 0 */
   /* :TODO: is this correct?? */
   MPI_Comm_rank (*(applyParams->mpiParams->comm), &myRank);

   if (myRank != 0)
   {
      ABORT (status, 
            STOCHASTICAPPLYSEARCHH_ERANK, 
            STOCHASTICAPPLYSEARCHH_MSGERANK);
   }

   if (params->currentSegment < params->initParams->numSegments)
   {

      /* we still have data to analyze, so carry on with search */


      /* 
       *
       * allocate temporary memory for the frequency data 
       *
       */

      fseries1.data = NULL;
      LALCCreateVector (status->statusPtr, 
			&(fseries1.data), 
			params->initParams->numPoints);
      CHECKSTATUSPTR (status);

      fseries2.data = NULL;
      LALCCreateVector (status->statusPtr, 
			&(fseries2.data), 
			params->initParams->numPoints);
      CHECKSTATUSPTR (status);

      ccSpectrum.data = NULL;
      LALCCreateVector (status->statusPtr, 
            &(ccSpectrum.data), 
            params->initParams->numFrequencies);
      CHECKSTATUSPTR (status);

      ccSpectrumOut.data = NULL;
      LALCCreateVector (status->statusPtr, 
            &(ccSpectrumOut.data), 
            params->initParams->outputNumFreq);
      CHECKSTATUSPTR (status);

      /* point dummySegment to the current segment to analyze */
      dummySegment = params->stochasticSegVec->data + params->currentSegment;

      // printf("%d \n", params->currentSegment);

      /* set fields of time interval */
      timeInterval.numberSamples = (UINT8) (params->initParams->numPoints);
      timeInterval.startSec = (UINT8) (dummySegment->data1->epoch.gpsSeconds);
      timeInterval.startNan 
	= (UINT8) (dummySegment->data1->epoch.gpsNanoSeconds);
      timeInterval.timeStepSize = (REAL8) (dummySegment->data1->deltaT);
      segStartNanoSeconds = 1000000000L * (INT8) (timeInterval.startSec);
      segStartNanoSeconds += timeInterval.startNan;
      segLengthNanoSeconds = (INT8) (1.0e9 * timeInterval.numberSamples * 
            timeInterval.timeStepSize);
      segStopNanoSeconds = segStartNanoSeconds + segLengthNanoSeconds;
      timeInterval.stopSec = (UINT8) (segStopNanoSeconds / 1000000000L);
      timeInterval.stopNan = (UINT8) (segStopNanoSeconds % 1000000000L);
      // printf("%d %d \n", timeInterval.startSec, timeInterval.startNan);


      /* 
       *
       * zero-pad and FFT the time-series data data 
       *
       */

      LALSZeroPadAndFFT (status->statusPtr,
            &fseries1,
            dummySegment->data1,
            params->fftPlan);
      CHECKSTATUSPTR (status);

      LALSZeroPadAndFFT (status->statusPtr,
            &fseries2,
            dummySegment->data2,
            params->fftPlan);
      CHECKSTATUSPTR (status);


      /* 
       *
       * build stochastic cross-correlation input structure 
       *
       */

      ccInput.hBarTildeOne  = &fseries1;
      ccInput.hBarTildeTwo  = &fseries2;
      ccInput.optimalFilter = params->optimalFilter;


      /* 
       *
       * calculate the cross-correlation spectrum 
       *
       */
      LALStochasticCrossCorrelationSpectrum (status->statusPtr,
					     &ccSpectrum,
					     &ccInput,
					     params->initParams->epochsMatch);
      CHECKSTATUSPTR (status);


      cgParams.length = params->initParams->outputNumFreq;
      cgParams.f0     = params->initParams->outputF0;
      cgParams.deltaF = params->initParams->outputDeltaF;

      /* Coarse-grain the output to the desired resolution */
      TRY( LALCCoarseGrainFrequencySeries(status->statusPtr,
					  &ccSpectrumOut,
					  &ccSpectrum,
					  &cgParams),
	   status);

      /* 
       *
       * build output for metadata database or trend frames 
       *
       */

      StochasticBuildOutput(status->statusPtr,
            output,
            &ccSpectrumOut,
            timeInterval);
      CHECKSTATUSPTR (status);


      /* 
       *
       * remove temporary memory for the frequency data 
       *
       */

      TRY( LALCDestroyVector (status->statusPtr, &(ccSpectrumOut.data)),
	   status );

      LALCDestroyVector (status->statusPtr, &(ccSpectrum.data));
      CHECKSTATUSPTR (status);

      LALCDestroyVector (status->statusPtr, &(fseries2.data));
      CHECKSTATUSPTR (status);

      LALCDestroyVector (status->statusPtr, &(fseries1.data));
      CHECKSTATUSPTR (status);

      /*
       *  
       * compute fraction of search remaining for wrapper and set 
       * output not finished 
       * 
       * :NOTE: this is what duncan did for simple.so even though
       * it would seem that i should increment to the pointer
       * to the current segment before computing fracRemaining and
       * before determining how to set notFinished.
       *
       */  

      output->fracRemaining = 1.0 - 
         (REAL4)(params->currentSegment)/ 
         (REAL4)(params->initParams->numSegments);

      output->notFinished = STOCHASTICH_TRUE;


      /* increment pointer to current segment */
      (params->currentSegment)++;

   }
   else
   {
      /* we are finished with the search */

      output->fracRemaining = 0;
      output->notFinished   = STOCHASTICH_FALSE;
   }


   /* normal exit */ 
   DETATCHSTATUSPTR (status);
   RETURN (status);
}

